using System;
using System.IO;
using System.Windows.Forms;
using System.Collections;
using System.Collections.Generic;
namespace HaloMap
{
    public enum MapTypes
    {
        Internal,
        MainMenu,
        SPShared,
        MPShared,
        Bitmaps
    }

    public class Map
    {
        public TagMeta.ItemType DisplayType;
        public TagMeta SelectedMeta;
        public string filePath;
        public FileStream FS;
        public BinaryReader BR;
        public BinaryWriter BW;
        public bool isOpen = false;
        public MapTypes openMapType;
        public int PrimaryMagic;
        public int SecondaryMagic;
        public MapHeaderInfo MapHeader;
        public IndexHeaderInfo IndexHeader;
        public ObjectIndexInfo MetaInfo;
        public FileNamesInfo FileNames;
        public StringsInfo Strings;
        public MiscFunctions Functions;
        public BSPContainer BSP;
        public UnicodeTableReader Unicode;
        public MetaContainers.ugh_ ugh;
        public HaloVersionEnum HaloVersion;
        public BitmapLibraryLayout BitmapLibary;

        public bool error = false;
        public Map()
        {
        }
        public Map(string path)
        {
            filePath = path;
            Functions = new MiscFunctions();
            OpenMap(MapTypes.Internal);
            BR.BaseStream.Position = 4;
            int version = BR.ReadInt32();
            switch (version)
            {
                case 609:
                    HaloVersion = HaloVersionEnum.HaloCE;
                    break;
                case 8:
                    HaloVersion = HaloVersionEnum.Halo2;
                    break;
                case 5:
                    HaloVersion = HaloVersionEnum.Halo1;
                    break;
                default:
                    MessageBox.Show("This is not a supported map type.");
                    error = true;
                    return;
            }
            LoadMap();
            CloseMap();

        }

        public void LoadMap()
        {
            switch (HaloVersion)
            {
                case HaloVersionEnum.Halo2:
                    MapHeader = new MapHeaderInfo(ref BR, HaloVersion);
                    IndexHeader = new IndexHeaderInfo(ref BR, this);
                    MetaInfo = new ObjectIndexInfo(ref BR, this);
                    FileNames = new FileNamesInfo(ref BR, this);
                    Strings = new StringsInfo(ref BR, this);
                    CloseMap();
                    Unicode = new UnicodeTableReader(this.filePath, this.MetaInfo.Offset[0]);

                    break;
                case HaloVersionEnum.HaloCE:
                    BitmapLibary = new BitmapLibraryLayout(this);
                    MapHeader = new MapHeaderInfo(ref BR, HaloVersion);
                    IndexHeader = new IndexHeaderInfo(ref BR, this);
                    MetaInfo = new ObjectIndexInfo(ref BR, this);
                    FileNames = new FileNamesInfo(ref BR, this);
                    this.MapHeader.fileNamesSize = FileNames.FileNameStringsSize;
                    this.MapHeader.offsetTofileNames = FileNames.FileNameStringsOffset;
                    break;
                case HaloVersionEnum.Halo1:
                    //BitmapLibary = new BitmapLibraryLayout(this);
                    MapHeader = new MapHeaderInfo(ref BR, HaloVersion);
                    IndexHeader = new IndexHeaderInfo(ref BR, this);
                    MetaInfo = new ObjectIndexInfo(ref BR, this);
                    FileNames = new FileNamesInfo(ref BR, this);
                    this.MapHeader.fileNamesSize = FileNames.FileNameStringsSize;
                    this.MapHeader.offsetTofileNames = FileNames.FileNameStringsOffset;
                    break;
            }
            DisplayType = TagMeta.ItemType.Reflexive;
        }


        public enum HaloVersionEnum
        {
            Halo2,
            HaloCE,
            Halo1
        }

        public void Sign()
        {
            if (HaloVersion != HaloVersionEnum.Halo2) { MessageBox.Show("Only Sign Halo 2 Maps"); return; }
            OpenMap(MapTypes.Internal);
            int size = (int)this.FS.Length - 2048;
            int times = size / 4;
            int result = 0;
            int x;
            this.BR.BaseStream.Seek(2048, SeekOrigin.Begin);
            for (x = 0; x < times; x++)
            {
                result = result ^ this.BR.ReadInt32();
            }
            this.BW.BaseStream.Seek(720, SeekOrigin.Begin);
            this.BW.Write(result);

            this.MapHeader.signature = result;
            CloseMap();
        }
        public void OpenMap(MapTypes type)
        {
            if (isOpen == true && type == this.openMapType) { return; }
            CloseMap();
            try
            {

                string temp = "";
                int x = 0;
                switch (type)
                {
                    case MapTypes.Internal:
                        temp = filePath;
                        break;
                    case MapTypes.MainMenu:
                        x = filePath.LastIndexOf('\\');
                        temp = filePath.Substring(0, x + 1) + "mainmenu.map";
                        break;
                    case MapTypes.MPShared:
                        x = filePath.LastIndexOf('\\');
                        temp = filePath.Substring(0, x + 1) + "shared.map";
                        break;
                    case MapTypes.SPShared:
                        x = filePath.LastIndexOf('\\');
                        temp = filePath.Substring(0, x + 1) + "single_player_shared.map";
                        break;
                    case MapTypes.Bitmaps:
                        x = filePath.LastIndexOf('\\');
                        temp = filePath.Substring(0, x + 1) + "bitmaps.map";
                        break;
                }

                FS = new FileStream(temp, FileMode.Open);
                BR = new BinaryReader(FS);
                BW = new BinaryWriter(FS);
                isOpen = true;
                openMapType = type;
            }
            catch
            {
                MessageBox.Show(type.ToString() + " - Map Not Found");
            }

        }
        public void BufferReadWrite(ref BinaryReader BR, ref BinaryWriter BW, int size)
        {
            int buffersize = (int)0xA00000;
            int currread = 0;
            do
            {
                int tempsize = (size - currread);
                int buffsize = buffersize;
                if (tempsize < buffersize)
                {
                    buffsize = size - currread;
                    BW.Write(BR.ReadBytes(buffsize));
                    break;
                }
                BW.Write(BR.ReadBytes(buffsize));
                currread += buffersize;
                GC.Collect(0);
                GC.WaitForPendingFinalizers();
            }
            while (currread != -1);

        }
        public void CloseMap()
        {
            if (isOpen == false) { return; }
            BR.Close();
            BW.Close();
            FS.Close();
            isOpen = false;
        }

    }

    public class BSPContainer
    {
        #region Fields
        public BSPInfo[] sbsp;
        public int bspcount;

        public class Palette_Color
        {
            public int r, g, b, a;
        }
        #endregion
        #region Constructor;
        public BSPContainer(int mapnumber)
        {

            switch (Maps.map[mapnumber].HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    //Halo2BSPContainer(mapnumber);             //REMOVED FOR H2V FOR CRASHING ISSUES
                    break;
                case Map.HaloVersionEnum.Halo1:
                    Halo1Container(mapnumber);
                    break;
                case Map.HaloVersionEnum.HaloCE:

                    HaloCEContainer(mapnumber);
                    break;
            }
        }
        public void HaloCEContainer(int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[0] + 1444;
            bspcount = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            sbsp = new BSPInfo[bspcount];
            for (int x = 0; x < bspcount; x++)
            {
                sbsp[x] = new BSPInfo
                {
                    pointerOffset = tempr + (x * 32) - Maps.map[mapnumber].MetaInfo.Offset[0]
                };
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 32);
                sbsp[x].offset = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].size = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].magic = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].offset;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 32) + 28;
                sbsp[x].ident = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].tagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].ident, mapnumber);
                Maps.map[mapnumber].MetaInfo.Offset[sbsp[x].tagnumber] = sbsp[x].offset;
                Maps.map[mapnumber].MetaInfo.Size[sbsp[x].tagnumber] = sbsp[x].size;
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].offset+4;
                sbsp[x].Halo1VerticeCount = Maps.map[mapnumber].BR.ReadInt32();

                sbsp[x].lightmapoffset = 24;
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].lightmapoffset + 12;

                sbsp[x].lightmapident  = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].lightmaptagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].ident, mapnumber);
            }
            Maps.map[mapnumber].CloseMap();
        }
        public void Halo1Container(int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[0] + 1444;
            bspcount = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].PrimaryMagic;
            sbsp = new BSPInfo[bspcount];
            for (int x = 0; x < bspcount; x++)
            {
                sbsp[x] = new BSPInfo
                {
                    pointerOffset = tempr + (x * 32) - Maps.map[mapnumber].MetaInfo.Offset[0]
                };
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 32);
                sbsp[x].offset = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].size = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].magic = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].offset;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 32) + 28;
                sbsp[x].ident = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].tagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].ident, mapnumber);
                Maps.map[mapnumber].MetaInfo.Offset[sbsp[x].tagnumber] = sbsp[x].offset;
                Maps.map[mapnumber].MetaInfo.Size[sbsp[x].tagnumber] = sbsp[x].size;
          
      
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].offset + 4;
                sbsp[x].Halo1VerticeCount = Maps.map[mapnumber].BR.ReadInt32();

                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].offset + 24 + (sbsp[x].Halo1VerticeCount * 12) + 4;
                int temprr = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].magic ;

                sbsp[x].lightmapoffset = temprr - sbsp[x].offset;
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].offset+sbsp[x].lightmapoffset + 12;

                sbsp[x].lightmapident = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].lightmaptagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].lightmapident, mapnumber);
            }
            Maps.map[mapnumber].CloseMap();
        }
        public void Halo2BSPContainer(int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Maps.map[mapnumber].BR.BaseStream.Position = Maps.map[mapnumber].MetaInfo.Offset[3] + 528;
            bspcount = Maps.map[mapnumber].BR.ReadInt32();
            int tempr = Maps.map[mapnumber].BR.ReadInt32() - Maps.map[mapnumber].SecondaryMagic;
            sbsp = new BSPInfo[bspcount];
            for (int x = 0; x < bspcount; x++)
            {
                sbsp[x] = new BSPInfo
                {
                    pointerOffset = tempr + (x * 68) - Maps.map[mapnumber].MetaInfo.Offset[3]
                };
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 68);
                sbsp[x].offset = Maps.map[mapnumber].BR.ReadInt32();
                Maps.map[mapnumber].Functions.ParsePointer(ref sbsp[x].offset, ref sbsp[x].location);
                sbsp[x].size = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].magic = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].offset;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 68) + 20;
                sbsp[x].ident = Maps.map[mapnumber].BR.ReadInt32();
                //	MessageBox.Show(sbsp[x].ident.ToString("X"));
                sbsp[x].tagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].ident, mapnumber);
                Maps.map[mapnumber].MetaInfo.Offset[sbsp[x].tagnumber] = sbsp[x].offset;
                Maps.map[mapnumber].MetaInfo.Size[sbsp[x].tagnumber] = sbsp[x].size;
                Maps.map[mapnumber].BR.BaseStream.Position = tempr + (x * 68) + 28;
                sbsp[x].lightmapident = Maps.map[mapnumber].BR.ReadInt32();
                sbsp[x].lightmaptagnumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(sbsp[x].lightmapident, mapnumber);
                 if (sbsp[x].lightmaptagnumber == -1) { continue; }
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].offset + 8;
                sbsp[x].lightmapoffset = Maps.map[mapnumber].BR.ReadInt32();
                if (sbsp[x].lightmapoffset == 0)
                {
                    sbsp[x].lightmapident = -1;
                    sbsp[x].lightmaptagnumber = -1;
                    if (MessageBox.Show("There is no lightmap for this bsp and the scenario is currently linked to a broken ID.\n Would you like Entity to fix it?", "Error", MessageBoxButtons.YesNo) == DialogResult.Yes)
                    {
                        Maps.map[mapnumber].BW.BaseStream.Position = tempr + (x * 68) + 28;
                        Maps.map[mapnumber].BW.Write(int.Parse("FFFFFFFF", System.Globalization.NumberStyles.HexNumber));
                    }

                    continue;
                }
                sbsp[x].lightmapoffset +=- sbsp[x].magic;
                sbsp[x].lightmapsize = sbsp[x].size + sbsp[x].offset - sbsp[x].lightmapoffset;
               
                Maps.map[mapnumber].MetaInfo.Offset[sbsp[x].lightmaptagnumber] = sbsp[x].lightmapoffset;
                Maps.map[mapnumber].MetaInfo.Size[sbsp[x].lightmaptagnumber] = sbsp[x].lightmapsize;
                ///light map bitmap
                Maps.map[mapnumber].BR.BaseStream.Position = sbsp[x].lightmapoffset + 128;
                int tempc = Maps.map[mapnumber].BR.ReadInt32();                                               //CRASHES HERE AFTER FIXING HEADERS
                int temprx = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].magic;
                Maps.map[mapnumber].BR.BaseStream.Position = temprx + 28;
                sbsp[x].LightMap_TagNumber = Maps.map[mapnumber].Functions.Meta.FindMetaByID(Maps.map[mapnumber].BR.ReadInt32(), mapnumber);
                ///light map palettes
                Maps.map[mapnumber].BR.BaseStream.Position = temprx + 8;
                int tempc2 = Maps.map[mapnumber].BR.ReadInt32();
                int tempr2 = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].magic;
                sbsp[x].palettesoffset = tempr2;
                for (int y = 0; y < tempc2; y++)
                {
                    Maps.map[mapnumber].BR.BaseStream.Position = tempr2 + (y * 1024);
                    Palette_Color[] pc = new Palette_Color[256];
                    for (int z = 0; z < 256; z++)
                    {
                        pc[z] = new Palette_Color
                        {
                            r = Maps.map[mapnumber].BR.ReadByte(),
                            g = Maps.map[mapnumber].BR.ReadByte(),
                            b = Maps.map[mapnumber].BR.ReadByte(),
                            a = Maps.map[mapnumber].BR.ReadByte()
                        };
                    }
                    sbsp[x].LightMap_Palettes.Add(pc);
                }

                Maps.map[mapnumber].BR.BaseStream.Position = temprx + 40;
                tempc2 = Maps.map[mapnumber].BR.ReadInt32();
                tempr2 = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].magic;
                sbsp[x].VisualChunk_Bitmap_Index = new int[tempc2];
                sbsp[x].VisualChunk_LightMap_Index = new int[tempc2];

                if (tempc2 != 0)
                {
                    Maps.map[mapnumber].BR.BaseStream.Position = tempr2;
                    for (int y = 0; y < tempc2; y++)
                    {
                        sbsp[x].VisualChunk_Bitmap_Index[y] = Maps.map[mapnumber].BR.ReadInt16();
                        sbsp[x].VisualChunk_LightMap_Index[y] = Maps.map[mapnumber].BR.ReadInt16();

                    }
                }

                Maps.map[mapnumber].BR.BaseStream.Position = temprx + 72;
                tempc2 = Maps.map[mapnumber].BR.ReadInt32();
                tempr2 = Maps.map[mapnumber].BR.ReadInt32() - sbsp[x].magic;
                sbsp[x].SceneryChunk_Bitmap_Index = new int[tempc2];
                sbsp[x].SceneryChunk_LightMap_Index = new int[tempc2];

                if (tempc2 > 0)
                {
                    Maps.map[mapnumber].BR.BaseStream.Position = tempr2;
                    for (int y = 0; y < tempc2; y++)
                    {
                        sbsp[x].SceneryChunk_Bitmap_Index[y] = Maps.map[mapnumber].BR.ReadInt16();
                        sbsp[x].SceneryChunk_LightMap_Index[y] = Maps.map[mapnumber].BR.ReadInt16();

                    }
                }
            }

            Maps.map[mapnumber].CloseMap();
        }
        public int FindBSPNumberByBSPIdent(int id)
        {
            for (int x = 0; x < sbsp.Length; x++)
            {
                if (sbsp[x].ident == id)
                {
                    return x;
                }
            }
            return -1;
        }
        public int FindBSPNumberByLightMapIdent(int id)
        {
            for (int x = 0; x < sbsp.Length; x++)
            {
                if (sbsp[x].lightmapident == id)
                {
                    return x;
                }
            }
            return -1;
        }
        #endregion
    }

    public class BSPInfo
    {
        #region Fields
        public int pointerOffset;
        public int offset;
        public int size;
        public int magic;
        public int ident;
        public int tagnumber;
        public MapTypes location;

        public int lightmapoffset;
        public int lightmaptagnumber;
        public int lightmapident;
        public int lightmapsize;


        public int LightMap_TagNumber=-1;

        public int palettesoffset;
        public List<BSPContainer.Palette_Color[]> LightMap_Palettes = new List<BSPContainer.Palette_Color[]>();
        public int[] VisualChunk_LightMap_Index;
        public int[] VisualChunk_Bitmap_Index;
        public int[] SceneryChunk_LightMap_Index;
        public int[] SceneryChunk_Bitmap_Index;

        public int Halo1VerticeCount;
        #endregion
        #region Constructor;
        public BSPInfo()
        {

        }
        #endregion
        public void WritePalettes(int mapnumber)
        {
            Maps.map[mapnumber].OpenMap(MapTypes.Internal);
            Maps.map[mapnumber].BW.BaseStream.Position = palettesoffset;
            foreach (BSPContainer.Palette_Color[] palette in LightMap_Palettes)
            {
                for (int x = 0; x < 256; x++)
                {
                    Maps.map[mapnumber].BW.Write((byte)palette[x].r);
                    Maps.map[mapnumber].BW.Write((byte)palette[x].g);
                    Maps.map[mapnumber].BW.Write((byte)palette[x].b);
                    Maps.map[mapnumber].BW.Write((byte)palette[x].a);
                }

            }
            Maps.map[mapnumber].CloseMap();
        }
    }


    public class MapHeaderInfo
    {
        public int fileSize;
        public int indexOffset;
        public int metaStart;
        public int metaSize;
        public int combinedSize;
        public int offsetToStringNames1;
        public int scriptReferenceCount;
        public int sizeOfScriptReference;
        public int offsetToStringIndex;
        public int offsetToStringNames2;
        public int offsetToCrazy;
        public int sizeOfCrazy;
        public string mapName;
        public string scenarioPath;
        public MapTypes mapType;
        public int fileCount;
        public int offsetTofileNames;
        public int fileNamesSize;
        public int offsetTofileIndex;
        public int signature;

        public MapHeaderInfo(ref BinaryReader BR, Map.HaloVersionEnum haloversion)
        {
            switch (haloversion)
            {
                case Map.HaloVersionEnum.Halo2:
                    LoadHalo2MapHeaderInfo(ref BR);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    LoadHalo2MapHeaderInfo(ref BR);
                    break;
            }
        }
        public void LoadHaloCEMapHeaderInfo(ref BinaryReader BR)
        {
            BR.BaseStream.Position = 8;
            fileSize = BR.ReadInt32();
            BR.BaseStream.Position = 16;
            indexOffset = BR.ReadInt32();
            metaSize = BR.ReadInt32();


        }
        public void LoadHalo2MapHeaderInfo(ref BinaryReader BR)
        {
            //map stuff
            BR.BaseStream.Position = 8;
            fileSize = BR.ReadInt32();
            BR.BaseStream.Position = 16;
                        //ORIGINAL indexOffset = BR.ReadInt32();
                        //ORIGINAL metaStart = BR.ReadInt32();
                        //ORIGINAL metaSize = BR.ReadInt32();
                        //ORIGINAL combinedSize = BR.ReadInt32();
                        //ORIGINAL BR.BaseStream.Position = 344;
                        //ORIGINAL offsetToCrazy = BR.ReadInt32();      //I THINK THIS IS FILE TABLE OFFSET??
                        //ORIGINAL sizeOfCrazy = BR.ReadInt32();
            indexOffset = BR.ReadInt32();
            metaStart = BR.ReadInt32();     //THIS IS ALSO OFFSET SIZE BTW
            metaSize = BR.ReadInt32();
            combinedSize = BR.ReadInt32();



            //string stuff
            BR.BaseStream.Position = 380;               // BR.BaseStream.Position = 352;
            offsetToStringNames1 = BR.ReadInt32();      //stringIDTableOffset    //NEVER USED
            BR.BaseStream.Position = 368;               //NEW
            scriptReferenceCount = BR.ReadInt32();      //stringIDCount
            sizeOfScriptReference = BR.ReadInt32();     //stringIDTableSize
            offsetToStringIndex = BR.ReadInt32();       //stringIDIndexOffset
            offsetToStringNames2 = BR.ReadInt32();      //stringIDTableOffset
            //map names and code to check if it is an external map
            BR.BaseStream.Position = 420;               // BR.BaseStream.Position = 408;
            mapName = new string(BR.ReadChars(36));
            BR.BaseStream.Position = 456;               // BR.BaseStream.Position = 444;
            scenarioPath = new string(BR.ReadChars(64));
            mapType = MapTypes.Internal;
            if (scenarioPath.IndexOf("scenarios\\ui\\mainmenu\\mainmenu") != -1) { mapType = MapTypes.MainMenu; }
            if (scenarioPath.IndexOf("scenarios\\shared\\shared") != -1) { mapType = MapTypes.MPShared; }
            if (scenarioPath.IndexOf("scenarios\\shared\\single_player_shared") != -1) { mapType = MapTypes.SPShared; }
            //read in stuff about meta names
            BR.BaseStream.Position = 716;               // BR.BaseStream.Position = 704;
            fileCount = BR.ReadInt32();                 //fileTableCount
            offsetTofileNames = BR.ReadInt32();         //fileTableOffset
            fileNamesSize = BR.ReadInt32();             //fileTableSize
            offsetTofileIndex = BR.ReadInt32();         //fileIndexTableOffset
            //signature
            signature = BR.ReadInt32();                 //H2V just has FF and bunch of 0's but idk
        }
    }

    public class IndexHeaderInfo
    {

        public int tagTypeCount;
        public int tagsOffset;
        public int metaCount;
        public int constant;


        public int ModelRawDataOffset;
        public int VerticeObjectCount;
        public int ModelIndicesOffset;
        public int IndicesObjectCount;
        public int ModelRawDataSize;

        public IndexHeaderInfo(ref BinaryReader BR, Map map)
        {
            switch (map.HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    LoadHalo2IndexHeaderInfo(ref BR, map);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                    LoadHaloCEIndexHeaderInfo(ref BR, map);
                    break;
                case Map.HaloVersionEnum.Halo1:
                    LoadHalo1IndexHeaderInfo(ref BR, map);
                    break;
            }
        }
        public void LoadHalo1IndexHeaderInfo(ref BinaryReader BR, Map map)
        {
            BR.BaseStream.Position = map.MapHeader.indexOffset;
            constant = BR.ReadInt32();
            map.PrimaryMagic = constant - (map.MapHeader.indexOffset + 36);

            BR.ReadInt32();
            BR.ReadInt32();
            metaCount = BR.ReadInt32();
            VerticeObjectCount = BR.ReadInt32();
            ModelRawDataOffset = BR.ReadInt32() - map.PrimaryMagic;
            IndicesObjectCount = BR.ReadInt32();
            ModelIndicesOffset = BR.ReadInt32() - map.PrimaryMagic;
            tagsOffset = map.MapHeader.indexOffset + 36;

        }
        public void LoadHaloCEIndexHeaderInfo(ref BinaryReader BR, Map map)
        {
            BR.BaseStream.Position = map.MapHeader.indexOffset;
            constant = BR.ReadInt32();
            map.PrimaryMagic = constant - (map.MapHeader.indexOffset + 40);

            BR.ReadInt32();
            BR.ReadInt32();
            metaCount = BR.ReadInt32();
            VerticeObjectCount = BR.ReadInt32();
            ModelRawDataOffset = BR.ReadInt32();
            IndicesObjectCount = BR.ReadInt32();
            ModelIndicesOffset = BR.ReadInt32();
            ModelRawDataSize = BR.ReadInt32();
            tagsOffset = map.MapHeader.indexOffset + 40;

        }
        public void LoadHalo2IndexHeaderInfo(ref BinaryReader BR, Map map)
        {
            BR.BaseStream.Position = map.MapHeader.indexOffset;
            constant = BR.ReadInt32();
            map.PrimaryMagic = constant - (map.MapHeader.indexOffset + 32);
            tagTypeCount = BR.ReadInt32();
            tagsOffset = BR.ReadInt32() - map.PrimaryMagic;
            BR.BaseStream.Position = tagsOffset + 8;
            map.SecondaryMagic = BR.ReadInt32() - (map.MapHeader.indexOffset + map.MapHeader.metaStart);

            BR.BaseStream.Position = map.MapHeader.indexOffset + 24;
            metaCount = BR.ReadInt32();
        }
    }


    public class ObjectIndexInfo
    {
        public Hashtable TagTypes;
        public int TagTypesCount;
        public string[] TagType;
        public string[] TagType2;
        public string[] TagType3;
        public int[] Ident;
        public int[] Offset;
        public int[] Size;
        public int lowident;
        public int highident;
        public Hashtable identHT;
        public int[] stringoffset;
        public bool[] external;
        public int[] bitmapindex;
        public ObjectIndexInfo(ref BinaryReader BR, Map map)
        {
            switch (map.HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    LoadHalo2ObjectIndexInfo(ref BR, map);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                    LoadHaloCEObjectIndexInfo(ref BR, map);
                    break;
                case Map.HaloVersionEnum.Halo1:
                    LoadHalo1ObjectIndexInfo(ref BR, map);
                    break;
            }
        }

        public void LoadHalo1ObjectIndexInfo(ref BinaryReader BR, Map map)
        {
            TagTypes = new Hashtable();
            identHT = new Hashtable();
            TagType = new string[map.IndexHeader.metaCount];
            TagType2 = new string[map.IndexHeader.metaCount];
            TagType3 = new string[map.IndexHeader.metaCount];
            Ident = new int[map.IndexHeader.metaCount];
            Offset = new int[map.IndexHeader.metaCount];
            Size = new int[map.IndexHeader.metaCount];
            stringoffset = new int[map.IndexHeader.metaCount];
            string[] temptagtypes = new string[500];
            external = new bool[map.IndexHeader.metaCount];
            bitmapindex = new int[map.IndexHeader.metaCount];
            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32);
                char[] tempchar = BR.ReadChars(4);
              
                    Array.Reverse(tempchar);
                  //  BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 4;
                 //   char[] tempchar2 = BR.ReadChars(4);
                //    Array.Reverse(tempchar2);
                //    BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 8;
                //    char[] tempchar3 = BR.ReadChars(4);
                 //   Array.Reverse(tempchar3);
              
       
                string tempstring = new string(tempchar);
              //  string tempstring2 = new string(tempchar2);
              //  string tempstring3 = new string(tempchar3);

     

                object tempobj = TagTypes[tempstring];
                if (tempobj == null)
                {
                    TagTypes.Add(tempstring, tempstring);
                }
                TagType[x] = tempstring;



                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 12;
                Ident[x] = BR.ReadInt32();
                identHT.Add(Ident[x], x);
                if (x == 0) { lowident = Ident[x]; }
                if (Ident[x] > highident) { highident = Ident[x]; }
                if (Ident[x] < lowident) { lowident = Ident[x]; }
                stringoffset[x] = BR.ReadInt32() - map.PrimaryMagic;
                Offset[x] = BR.ReadInt32();


                Offset[x] -= map.PrimaryMagic;


            }

            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {


                int end = map.MapHeader.fileSize;

                for (int xx = 0; xx < map.IndexHeader.metaCount; xx++)
                {
                    if (Offset[xx] < end && Offset[xx] > Offset[x] && external[xx] != true && Offset[xx] > 0) end = Offset[xx];
                }
                Size[x] = end - Offset[x];
                if (Size[x] < 0) { Size[x] = 0; }




            }
        }
        public void LoadHaloCEObjectIndexInfo(ref BinaryReader BR, Map map)
        {
            TagTypes = new Hashtable();
            identHT = new Hashtable();
            TagType = new string[map.IndexHeader.metaCount];
            TagType2 = new string[map.IndexHeader.metaCount];
            TagType3 = new string[map.IndexHeader.metaCount];
            Ident = new int[map.IndexHeader.metaCount];
            Offset = new int[map.IndexHeader.metaCount];
            Size = new int[map.IndexHeader.metaCount];
            stringoffset = new int[map.IndexHeader.metaCount];
            string[] temptagtypes = new string[500];
            external = new bool[map.IndexHeader.metaCount];
            bitmapindex = new int[map.IndexHeader.metaCount];
            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32);
                char[] tempchar = BR.ReadChars(4);
                Array.Reverse(tempchar);
                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 4;
                char[] tempchar2 = BR.ReadChars(4);
                Array.Reverse(tempchar2);
                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 8;
                char[] tempchar3 = BR.ReadChars(4);
                Array.Reverse(tempchar3);
                string tempstring = new string(tempchar);
                string tempstring2 = new string(tempchar2);
                string tempstring3 = new string(tempchar3);



                object tempobj = TagTypes[tempstring];
                if (tempobj == null)
                {
                    TagTypes.Add(tempstring, tempstring);
                }
                TagType[x] = tempstring;



                BR.BaseStream.Position = map.IndexHeader.tagsOffset + (x * 32) + 12;
                Ident[x] = BR.ReadInt32();
                identHT.Add(Ident[x], x);
                if (x == 0) { lowident = Ident[x]; }
                if (Ident[x] > highident) { highident = Ident[x]; }
                if (Ident[x] < lowident) { lowident = Ident[x]; }
                stringoffset[x] = BR.ReadInt32() - map.PrimaryMagic;
                Offset[x] = BR.ReadInt32();

                if (TagType[x] == "bitm" && map.BitmapLibary.error == false)
                {

                    if (Offset[x] < map.PrimaryMagic)
                    {
                        int tempindex = Offset[x];
                        bitmapindex[x] = tempindex;
                        Size[x] = map.BitmapLibary.RawSize[tempindex];
                        Offset[x] = map.BitmapLibary.RawOffset[tempindex];
                        external[x] = true;
                    }
                    else
                    {
                        Offset[x] -= map.PrimaryMagic;
                    }
                }
                else if (TagType[x] == "bitm" && map.BitmapLibary.error == true)
                {
                    if (Offset[x] < map.PrimaryMagic)
                    {

                        Size[x] = 0;
                        Offset[x] = 0;
                    }
                    else
                    {
                        Offset[x] -= map.PrimaryMagic;
                    }
                }
                else
                {
                    Offset[x] -= map.PrimaryMagic;
                }

            }

            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {


                int end = map.MapHeader.fileSize;
                if (external[x] != true)
                {
                    for (int xx = 0; xx < map.IndexHeader.metaCount; xx++)
                    {
                        if (Offset[xx] < end && Offset[xx] > Offset[x] && external[xx] != true && Offset[xx] > 0) end = Offset[xx];
                    }
                    Size[x] = end - Offset[x];
                    if (Size[x] < 0) { Size[x] = 0; }


                }

            }
        }
        public void LoadHalo2ObjectIndexInfo(ref BinaryReader BR, Map map)
        {
            TagTypes = new Hashtable();
            identHT = new Hashtable();
            TagType = new string[map.IndexHeader.metaCount];
            Ident = new int[map.IndexHeader.metaCount];
            Offset = new int[map.IndexHeader.metaCount];
            Size = new int[map.IndexHeader.metaCount];
            string[] temptagtypes = new string[500];
            BR.BaseStream.Position = map.IndexHeader.tagsOffset;
            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                char[] tempchar = BR.ReadChars(4);
                Array.Reverse(tempchar);
                string tempstring = new string(tempchar);
                object tempobj = TagTypes[tempstring];
                if (tempobj == null)
                {
                    TagTypes.Add(tempstring, tempstring);
                }
                TagType[x] = tempstring;

                Ident[x] = BR.ReadInt32();
                identHT.Add(Ident[x], x);
                if (x == 0) { lowident = Ident[x]; }
                if (Ident[x] > highident) { highident = Ident[x]; }
                if (Ident[x] < lowident) { lowident = Ident[x]; }

                Offset[x] = BR.ReadInt32() - map.SecondaryMagic;
                Size[x] = BR.ReadInt32();
            }
            //TagTypes=new string[TagTypesCount];
            //Array.Copy(temptagtypes,0,TagTypes,0,TagTypesCount);
        }

    }



    public class FileNamesInfo
    {
        public string[] Name;
        public int[] Length;
        public int[] Offset;
        public int FileNameStringsSize;
        public int FileNameStringsOffset;
        public FileNamesInfo(ref BinaryReader BR, Map map)
        {

            switch (map.HaloVersion)
            {
                case Map.HaloVersionEnum.Halo2:
                    Halo2FileNamesInfo(ref BR, map);
                    break;
                case Map.HaloVersionEnum.HaloCE:
                case Map.HaloVersionEnum.Halo1:
                    HaloCEFileNamesInfo(ref BR, map);
                    break;
            }
        }
        public void HaloCEFileNamesInfo(ref BinaryReader BR, Map map)
        {
            Name = new string[map.IndexHeader.metaCount];
            Length = new int[map.IndexHeader.metaCount];
            Offset = new int[map.IndexHeader.metaCount];
            ///	Reads In Offsets Of	Meta Names
            BR.BaseStream.Position = map.MapHeader.offsetTofileIndex;
            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                Offset[x] = map.MetaInfo.stringoffset[x];
            }
            if (map.HaloVersion == Map.HaloVersionEnum.Halo1)
            {

                FileNameStringsOffset = map.MapHeader.indexOffset + 36 + (map.IndexHeader.metaCount * 32);
                FileNameStringsSize = map.IndexHeader.ModelRawDataOffset - FileNameStringsOffset;
            }
            else
            {
                FileNameStringsOffset = (map.MapHeader.indexOffset + 40 + (map.IndexHeader.metaCount * 32));
                FileNameStringsSize = map.MetaInfo.Offset[0] - FileNameStringsOffset;
            }

            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {

                //figures out length of string
                int len;
                if (x != map.IndexHeader.metaCount - 1)
                {
                    len = Offset[x + 1] - Offset[x];
                }
                else
                {
                    len = FileNameStringsOffset + FileNameStringsSize - Offset[x];
                }
                Length[x] = len - 1;
                ///	Reads in string
                BR.BaseStream.Position = Offset[x];
                Name[x] = new string(BR.ReadChars(len - 1));

            }
        }


        public void Halo2FileNamesInfo(ref BinaryReader BR, Map map)
        {
            Name = new string[map.IndexHeader.metaCount];
            Length = new int[map.MapHeader.fileCount];
            Offset = new int[map.MapHeader.fileCount];
            ///	Reads In Offsets Of	Meta Names
            BR.BaseStream.Position = map.MapHeader.offsetTofileIndex;
            for (int x = 0; x < map.MapHeader.fileCount; x++)
            {
                Offset[x] = BR.ReadInt32();
            }

            for (int x = 0; x < map.IndexHeader.metaCount; x++)
            {
                if (x < map.MapHeader.fileCount)
                {
                    //figures out length of string
                    int len;
                    if (x != map.MapHeader.fileCount - 1)
                    {
                        len = Offset[x + 1] - Offset[x];
                    }
                    else
                    {
                        len = map.MapHeader.fileNamesSize - Offset[x];
                    }
                    Length[x] = len - 1;
                    ///	Reads in string
                    BR.BaseStream.Position = map.MapHeader.offsetTofileNames + Offset[x];
                    Name[x] = new string(BR.ReadChars(len - 1));
                }
                else
                {

                    int tempint = x - map.MapHeader.fileCount;
                    Name[x] = "Unknown Meta (" + tempint.ToString() + ")";
                }
            }
        }
    }

    public class StringsInfo
    {
        public string[] Name;
        public int[] Length;
        public int[] Offset;

        public StringsInfo(ref BinaryReader BR, Map map)
        {
            Name = new string[map.MapHeader.scriptReferenceCount];
            Length = new int[map.MapHeader.scriptReferenceCount];
            Offset = new int[map.MapHeader.scriptReferenceCount];
            ///	Reads In Offsets Of	Meta Names
            BR.BaseStream.Position = map.MapHeader.offsetToStringIndex;
            for (int x = 0; x < map.MapHeader.scriptReferenceCount; x++)
            {
                Offset[x] = BR.ReadInt32();
            }

            for (int x = 0; x < map.MapHeader.scriptReferenceCount; x++)
            {
                //figures out length of string
                int len;
                if (x != map.MapHeader.scriptReferenceCount - 1)
                {
                    len = Offset[x + 1] - Offset[x];
                }
                else
                {
                    len = map.MapHeader.sizeOfScriptReference - Offset[x];
                }
                Length[x] = len - 1;
                ///	Reads in string
                BR.BaseStream.Position = map.MapHeader.offsetToStringNames2 + Offset[x];
                Name[x] = new string(BR.ReadChars(len - 1));
            }
        }
    }



    public class BitmapLibraryLayout
    {
        public string[] Name;
        public int[] StringLength;
        public int[] StringOffset;
        public int[] MetaOffset;
        public int[] MetaSize;
        public int[] RawOffset;
        public int[] RawSize;
        public int BitmapCount;
        public int FileNameStringsOffset;
        public int BitmapIndexOffset;
        public int FileSize;
        public bool error = false;
        public BitmapLibraryLayout(Map map)
        {

            switch (map.HaloVersion)
            {
                case Map.HaloVersionEnum.HaloCE:
                    //try
                    //   {
                    Map m = new Map
                    {
                        filePath = map.filePath
                    };
                    m.OpenMap(MapTypes.Bitmaps);
                    HaloCEBitmapLibraryLayout(ref m.BR);
                    m.CloseMap();
                    //    }
                    //    catch
                    //   {
                    //    error = true;
                    //   }
                    break;
            }
        }
        public void HaloCEBitmapLibraryLayout(ref BinaryReader BR)
        {
            FileSize = (int)BR.BaseStream.Length;
            BR.BaseStream.Position = 4;
            FileNameStringsOffset = BR.ReadInt32();
            BitmapIndexOffset = BR.ReadInt32();
            BitmapCount = BR.ReadInt32();
            StringLength = new int[BitmapCount];
            StringOffset = new int[BitmapCount];
            MetaOffset = new int[BitmapCount];
            MetaSize = new int[BitmapCount];
            RawOffset = new int[BitmapCount];
            RawSize = new int[BitmapCount];
            BR.BaseStream.Position = BitmapIndexOffset;
            for (int x = 0; x < BitmapCount; x++)
            {
                StringOffset[x] = BR.ReadInt32();
                RawSize[x] = BR.ReadInt32();
                RawOffset[x] = BR.ReadInt32();
                if (x != 0)
                {
                    StringLength[x - 1] = StringOffset[x] - StringOffset[x - 1] - 1;
                    if (x == BitmapCount - 1)
                    {
                        StringLength[x] = BitmapIndexOffset - StringOffset[x] - 1;
                    }
                }

            }
            
            Name = new string[BitmapCount];
            for (int x = 0; x < BitmapCount; x++)
            {
                BR.BaseStream.Position = FileNameStringsOffset + StringOffset[x];
                Name[x] = new string(BR.ReadChars(StringLength[x]));
            }


        }

    }



}